/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file conn_mgr.cpp
 *
 * Implements the cconn_mgr class, which is a manager (and Abstract Factory)
 * for the cconn class. For more information, see conn_mgr.h.
 *
 * TODO:
 * -call srand() somewhere
 * -some methods return conns that could be deleted before return call finished
 * -unify eclosed and efailed, and streamline status stuff
 * -find memory bug that keeps us from erasing closed conns
 * -sending data msgs to self is valid, but right now we return failed instance
 * -move p4_get_num_conns and p4_get_conns to p4client.cpp
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "conn_mgr.h"
#include "msg_data.h"
#include "msg_lka.h"
#include "msg_connect.h"
#include "receiveq.h"
#include "msg_factory.h"
#include "lka.h"
#include "debug.h"
#include "settings.h"
#include "sys.h"
#include "p4client.h"
#include "p4error.h"

#ifdef _WIN32
#include <Winsock2.h>
#else
#include <sys/socket.h>
#include <sys/types.h> // for FD_SET
#include <arpa/inet.h> // for inet_addr
#include <netdb.h>     // for gethostbyname
#include <unistd.h>    // for gethostname
#endif

#include <cassert>
#include <cstdlib>
#include <cerrno>

extern string glocal_user;

/*****************************************************************************
 * INITIALIZE STATIC MEMBERS
 *****************************************************************************/
const uint cconn_mgr::kideal_min_conns
  = csettings::instance().get<uint>("ideal_min_conns");

const uint cconn_mgr::kideal_max_conns
  = csettings::instance().get<uint>("ideal_max_conns");

const uint cconn_mgr::kselect_timeout_msecs
  = csettings::instance().get<uint>("select_timeout_msecs");

const float cconn_mgr::kdiscover_prob
  = csettings::instance().get<float>("discover_prob");

const ushort cconn_mgr::kp4_port
  = csettings::instance().get<ushort>("p4_port");


/*****************************************************************************
 * CLIENT API METHODS
 *****************************************************************************/
uint p4_get_num_conns(int *num_conns)
{
  cconn_mgr::instance().client_get_num_conns(*num_conns);
  return kerr_success;
}

uint p4_get_conns(p4_conn_info *conns, int num_conns, int *num_written)
{
  cconn_mgr::instance().client_get_conns(conns, num_conns, *num_written);
  return kerr_success;
}


/*****************************************************************************
 * PUBLIC METHODS
 *****************************************************************************/
uint cconn_mgr::receive()
{
  cdebug::instance() << acquire("conn_mgr_receive_manage")
                     << "receive()"
                     << release();
  fd_set conn_fds;
#ifdef _WIN32
  int remote_size;
#else
  uint remote_size;
#endif

  int max_fd = 0;
  FD_ZERO(&conn_fds);

  pthread_mutex_lock(&mtx);


  // populate conn_fds (the fd_set) with connections
  list<cconn>::const_iterator it;
  for (it = conns.begin(); it != conns.end(); it++) {
    cconn::estatus_t stat = it->get_status();
    if (stat == cconn::efailed || stat == cconn::eopening ||
      stat == cconn::eclosed)  // don't add bad cconns
      continue;

    int sock = it->get_socket();
    verify_socket(sock);

    FD_SET(sock, &conn_fds);  // add cconn's socket to fd_set
    if (sock > max_fd)
      max_fd = sock;
  }


  // populate conn_fds with listener socket
  FD_SET(listener, &conn_fds);
  if (listener > max_fd)
    max_fd = listener;


  // set timeout
  struct timeval timeout;
  timeout.tv_sec = cconn_mgr::kselect_timeout_msecs / 1000;
  timeout.tv_usec = (cconn_mgr::kselect_timeout_msecs % 1000) * 1000;
  assert(timeout.tv_sec <= 1);


  // select
  int ret = select(max_fd + 1, &conn_fds, 0, 0, &timeout);
  if (ret == -1) {
    cdebug::instance() << acquire("conn_mgr_debug")
                       << "select failed with errno " << errno
                       << release();
    pthread_mutex_unlock(&mtx);
    return kerr_select_failed;
  }


  // accept an incoming connection
  if (FD_ISSET(listener, &conn_fds)) {
    cdebug::instance() << acquire("conn_mgr_debug")
                       << "There's an incoming connection!"
                       << release();
    struct sockaddr_in remote;
    remote_size = sizeof(remote);

    int newsock = accept(listener,
                         reinterpret_cast<struct sockaddr *>(&remote),
                         &remote_size);

    if (newsock >= 0) {  // successfully accepted connection
      verify_socket(newsock);
      conns.push_front(cconn(newsock));
    } else
      cdebug::instance() << acquire("conn_mgr_debug")
                         << "failed accept, errno " << errno
                         << release();
  }


  // receive on current connections
  // 
  // iterate from end to beginning because we might kill connections, and the
  // iterator shouldn't be invalidated
  list<cconn>::reverse_iterator rit;
  for (rit = conns.rbegin(); rit != conns.rend(); rit++) {

    if (FD_ISSET(rit->get_socket(), &conn_fds)) {
      uint ret = rit->receive();  // make the conn receive
      if (ret != kerr_success) {  // error receiving, kill conn
        cdebug::instance() << acquire("conn_mgr_debug")
                           << "error " << ret << " in receiving"
                           << release();
        rit->close();

        // this causes problems. maybe a double-free somewhere?
//        conns.erase(rit.base());
      }
    }
  }

  pthread_mutex_unlock(&mtx);
  return kerr_success;
}


void cconn_mgr::manage()
{
  cdebug::instance() << acquire("conn_mgr_receive_manage")
                     << "manage()"
                     << release();

  pthread_mutex_lock(&mtx);

  uint num = num_connected();
  
  // are there too few connections?
  if (num < kideal_min_conns) {
    float randf = static_cast<float>(rand()) / RAND_MAX;

    if (randf < kdiscover_prob) {  // send CONNECT to random user
      cmsg_connect connect;
      string user;
      
      // if we have as many conns as entries in the lka, noop
      uint ret = clka::instance().random_user(conns.size() + 1, user);
      if (ret != kerr_success) {
        cdebug::instance() << acquire("conn_mgr_debug")
                           << "wanted to make new connection, but failed for "
                           << "reason  " << ret
                           << release();
      } else {
        if (get_conn(user).get_status() == cconn::efailed) {
          cdebug::instance() << acquire("conn_mgr_debug")
                             << "too few connections.  connecting to " << user
                             << release();
          connect.send(user, false);
        } else {
          cdebug::instance() << acquire("conn_mgr_debug")
                             << "too few connections, but already connected "
                             << "to " << user
                             << release();
        }
      }
    }

  // are there too many connections?
  } else if (num > kideal_max_conns) {
    list<cconn>::iterator it = random_conn();
    assert(it != conns.end());
    if (it->is_idle()) {            // idle connection, close it!
      it->close();
      // this causes problems. maybe a double-free somewhere?
//      conns.erase(it);
    }
  }

  pthread_mutex_unlock(&mtx);
}


uint cconn_mgr::listen()
{
  cdebug::instance() << acquire("conn_mgr_debug") << "listen()" << release();

  struct sockaddr_in local;
  memset(&(local.sin_zero), '\0', sizeof(local.sin_zero));

  pthread_mutex_lock(&mtx);

  // fill in local address
  uint ret = get_local_addr(local);
  if (ret != kerr_success) {
          pthread_mutex_unlock(&mtx);
    return ret;
  }
  
  // create and bind socket
  if ((listener = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    pthread_mutex_unlock(&mtx);
    return kerr_couldnt_create_socket;
  }

  if (bind(listener, reinterpret_cast<struct sockaddr *>(&local),
           sizeof(local)) < 0) {
    pthread_mutex_unlock(&mtx);
    return kerr_couldnt_bind_socket;
  }

#ifndef _WIN32
  // TODO: port to windows
  // TODO: lose the pesky "Address already in use" error message
  int yes = 1;
  if (setsockopt(listener, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) < 0) {
    pthread_mutex_unlock(&mtx);
    return kerr_couldnt_bind_socket;
    exit(1);
  } 
#endif


  // listen for incoming connections
  if (::listen(listener, 20) < 0) {    // queue up to 20 connections
      pthread_mutex_unlock(&mtx);
      return kerr_couldnt_listen;
  }

  // NOTE: incoming connections are accepted in the receive() method
  pthread_mutex_unlock(&mtx);
  return kerr_success;
}



cconn &cconn_mgr::get_conn(const string &username)
{
  cdebug::instance() << acquire("conn_mgr_debug")
                     << "get_conn() checking for existing connections to "
                     << username << "..."
                     << release();

  // do we already have a connection to this address?
  for (list<cconn>::iterator it = conns.begin(); it != conns.end(); it++) {
    if (username == it->peer_username && it->get_status() != cconn::efailed &&
        it->get_status() != cconn::eclosed) {

      // found open conn with that addr
      cdebug::instance() << "found!" << release();
      return *it;
    }
  }

  cdebug::instance() << "none found!" << release();
  return cconn::failed_instance();
}  


// 
// these implement the client api functions for connections
// 
void cconn_mgr::client_get_num_conns(int &num_conns) const
{
  num_conns = conns.size();
}

void cconn_mgr::client_get_conns(p4_conn_info *conn_arr, int num_conns,
                                 int &num_written) const
{
  list<cconn>::const_iterator cit;
  int i;

  for (cit = conns.begin(), i = 0;
       cit != conns.end() && i < num_conns;
       cit++, i++) {

    // copy username and status
    assert(cit->peer_username.length() < kp4_username_len);
    cit->peer_username.copy(conn_arr[i].username, kp4_username_len);
    conn_arr[i].username[cit->peer_username.length()] = '\0';
    conn_arr[i].status = static_cast<p4_conn_status>(
      static_cast<int>(cit->status));

    // copy addr and port
    // TODO: this should be done with the lka, but it uses pretty loose
    // consistency, so i'm not sure it's exactly what we want here
    struct sockaddr_in sa;
#ifdef _WIN32
    int sa_size = sizeof(sockaddr_in);
#else
    uint sa_size = sizeof(sockaddr_in);
#endif

    if (getpeername(cit->sock, reinterpret_cast<sockaddr *>(&sa), &sa_size)
        < 0) {

      strcpy(conn_arr[i].addr, "255.255.255.255");
      conn_arr[i].port = 0;

    } else {
      string addr_str(inet_ntoa(sa.sin_addr));
      assert(addr_str.length() < kmax_ipstr_len);
      addr_str.copy(conn_arr[i].addr, kmax_ipstr_len);
      conn_arr[i].addr[addr_str.length()] = '\0';

      conn_arr[i].port = ntohs(sa.sin_port);
    }
  }

  num_written = i;
}


/*****************************************************************************
 * PROTECTED METHODS
 *****************************************************************************/
cconn_mgr::cconn_mgr()
{
  pthread_mutex_init(&mtx, NULL);
}


cconn_mgr::~cconn_mgr()
{
  cdebug::instance() << acquire("conn_mgr_debug") << "dtor()" << release();

  pthread_mutex_lock(&mtx);

  // this really would be a good thing to do here - since nothing else should
  // ever close the listener socket - but it throws an assert on windoze :(
//  verify_socket(listener);

#ifdef _WIN32
  closesocket(listener);
#else
  ::close(listener);
#endif


  // close all connections
  list<cconn>::iterator it;
  for (it = conns.begin(); it != conns.end(); it++)
    it->close();

  // connections will be *destroyed* later by list dtor

  pthread_mutex_unlock(&mtx);
  pthread_mutex_destroy(&mtx);
}


// default value for lock is true
const cconn &cconn_mgr::send(cmsg &msg, const string &username, bool lock)
{
  assert(username != glocal_user);  // local routing is done in cmsg::send

  if (lock)
    pthread_mutex_lock(&mtx);

  // is there already a conn to that user?
  cconn &conn = get_conn(username);
  if (conn.get_status() != cconn::efailed) {  // yes, send!
    uint ret = conn.send(msg);
    if (lock)
      pthread_mutex_unlock(&mtx);
    if (ret == kerr_success)
      return conn;
    else
      return cconn::failed_instance();
  }


  // look it up in the lka.  not there?  error
  ulong addr;
  ushort port;

  uint ret = clka::instance().get_address(username.c_str(), addr, port);
  if (ret != kerr_success) {
    if (lock)
      pthread_mutex_unlock(&mtx);
    return cconn::failed_instance();
  }
  
  // need to open a new conn. give it the message to send once it's open
  cconn new_conn(addr, port, msg, username.c_str());
  if (new_conn.get_status() == cconn::efailed) {
    cdebug::instance() << acquire("conn_mgr_debug")
                       << "send() failed to make new connection"
                       << release();
    if (lock)
      pthread_mutex_unlock(&mtx);
    return cconn::failed_instance();
  }

  // succeeded! return the newly created conn
  conns.push_front(new_conn);
  cconn &front = conns.front();
 
  if (lock)
    pthread_mutex_unlock(&mtx);
  return front;
}



/**
 * don't acquire mutex because we already have it.
 */
const cconn &cconn_mgr::random_send(cmsg &msg, const string &except_user)
{
  cdebug::instance() << acquire("conn_mgr_debug")
                     << "random_send " << msg << " except " << except_user
                     << release();

  // make sure there are connections open to people other than except_user
  // (otherwise, no one to send to, so break out early)
  uint num = num_connected();
  if (num == 0 ||
      (num == 1 && get_conn(except_user).get_status() == cconn::econnected))
    return cconn::failed_instance();

  list<cconn>::iterator it;
  do {  // get a random conn except for the one to username
    it = random_conn();
    assert(it != conns.end());
  } while (except_user == it->peer_username);
  it->send(msg);

  return *it;
}


// lock's default value is true
uint cconn_mgr::broadcast(cmsg &msg, bool lock)
{
  cdebug::instance() << acquire("conn_mgr_debug")
                     << "broadcast() " << msg
                     << release();

  if (num_connected() == 0)
    return kerr_no_connections;

  if (lock)
    pthread_mutex_lock(&mtx);

  // send on all connections except "except"
  // 
  // iterate from end to beginning because we might kill connections, and the
  // iterator shouldn't be invalidated
  bool sent = false;
  list<cconn>::reverse_iterator rit;

  for (rit = conns.rbegin(); rit != conns.rend(); rit++) {
    uint ret = rit->send(msg);
    if (ret != kerr_success) {    // send error, close connection
      rit->close();
      // this causes problems. maybe a double free somewhere?
//      conns.erase(rit.base());

    } else
      sent = true;
  }

  cdebug::instance() << acquire("conn_mgr_debug")
                     << "broadcast() finished"
                     << release();
  pthread_mutex_unlock(&mtx);
  return (sent) ? kerr_success : kerr_not_sent; 
}

/**
 * don't acquire mutex because we already have it.
 *     conn_mgr.receive()==>conn.receive()==>msg.process()==>
 *           msg_connect==>virt_process()==>conn_mgr.make_conn_permanent()
 */
void cconn_mgr::make_conn_permanent(const string &username)
{
  cdebug::instance() << acquire("conn_mgr_debug")
                     << "make_conn_permanent() " << username
                     << release();

  cconn &conn = get_conn(username);
  assert(conn.get_status() != cconn::efailed);

  conn.make_permanent();
}


/*****************************************************************************
 * PRIVATE METHODS AND HELPERS
 *****************************************************************************/
// don't acquire the mutex because this is called from methods that have
// already acquired it
uint cconn_mgr::num_connected() const
{
  uint num = 0;

  list<cconn>::const_iterator cit;
  for (cit = conns.begin(); cit != conns.end(); cit++)
    if (cit->get_status() == cconn::econnected)
      num++;

  return num;
}


// don't acquire the mutex because this is called from methods that have
// already acquired it
list<cconn>::iterator cconn_mgr::random_conn()
{
  // this is a signed int so i can compare to rand()
  int num = static_cast<int>(num_connected());

  if (num == 0)
    return conns.end();

  // no random access in lists, iterate and return elem w/random prob
  list<cconn>::iterator it = conns.begin();

  while (it->get_status() != cconn::econnected || rand() > RAND_MAX / num) {
    it++;
    if (it == conns.end())
      it = conns.begin();
  }

  return it;
}



/** get_local_addr
 *
 * Returns the local IP address in in_addr form (in *host* byte order). Returns
 * kerr_success on success, or the specific error otherwise.
 */
uint get_local_addr(struct sockaddr_in &sin)
{
  cdebug::instance() << acquire("conn_mgr_debug")
                     << "get_local_addr() on port " << cconn_mgr::kp4_port
                     << release();

  char local_name[256];
  struct hostent *host;

  // get local address
  gethostname(local_name, sizeof(local_name));
  local_name[255] = '\0';
  host = gethostbyname(local_name);

  if (!host) // gethostbyname should work with name or ip, right?
    return kerr_couldnt_create_socket;

  if (host->h_length != sizeof(sin.sin_addr.s_addr))
    return kerr_couldnt_create_socket;
  
  memcpy(&sin.sin_addr.s_addr, host->h_addr_list[0],
         sizeof(sin.sin_addr.s_addr));

  sin.sin_family = AF_INET;
  sin.sin_port = htons(cconn_mgr::kp4_port);

  return kerr_success;
}


/** resolve_addr
 *
 * Resolves a hostname (e.g. blaize.stanford.edu) or IP address (e.g.
 * 128.12.75.93) to an in_addr. If the given address is invalid, returns
 * INADDR_NONE.
 */
struct in_addr resolve_addr(const char *addr)
{
  struct hostent *host;
  struct in_addr sin;

  if ((host = gethostbyname(addr))) {
    // addr is a valid hostname
    memcpy(&sin.s_addr, host->h_addr_list[0], sizeof(sin.s_addr));
    cdebug::instance() << acquire("conn_mgr_debug")
                       << "resolved to " << inet_ntoa(sin)
                       << release();
    return sin;
 
  } else {
    // addr is an IP address (or if not, this will return INADDR_NONE)
    sin.s_addr = inet_addr(addr);
    return sin;
  }
}


/** verify_socket
 *
 * Checks the validity of the given socket. If it's not valid, verify_socket
 * prints an error message and exits.
 */
void verify_socket(int sock)
{
  struct sockaddr ret_addr;
#ifdef _WIN32
  int sa_size = sizeof(sockaddr);
#else
  uint sa_size = sizeof(sockaddr);
#endif

  if (getsockname(sock, &ret_addr, &sa_size) < 0) {
    cdebug::instance() << acquire("conn_mgr_debug")
                       << "verify_socket: socket " << sock
                       << " is invalid! errno " << errno
                       << release();

    // TODO: why is this commented out? grr...
    // assert(false);
  }
}

